/*

sudosh - sudo shell that supports input and output logging to syslog

Copyright 2004 and $Date: 2004/10/25 05:18:29 $
		Douglas Richard Hanks Jr.

Licensed under the Open Software License version 2.0

This program is free software; you can redistribute it
and/or modify it under the terms of the Open Software
License, version 2.0 by Lauwrence E. Rosen.

This program is distributed in the hope that it will be
useful, but WITHOUT ANY WARRANTY; without even the implied
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
PURPOSE.  See the Open Software License for details.

*/

#include <errno.h>
#include <stdio.h>
#include <stdarg.h>
#include <unistd.h>
#include <signal.h>
#include <pwd.h>

#include "config.h"

#ifdef TIME_WITH_SYS_TIME
#include <time.h>
#include <sys/time.h>
#else
 #ifdef HAVE_TIME_H
 #include <time.h>
 #endif
 #ifdef HAVE_SYS_TIME_H
 #include <sys/time.h>
 #endif
#endif

#ifdef HAVE_PTY_H
#include <pty.h>
#endif

#ifdef HAVE_FCNTL_H
#include <fcntl.h>
#endif

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_SYSLOG_H
#include <syslog.h>
#endif

#ifdef HAVE_SYS_STAT_H
#include <sys/stat.h>
#endif

#ifdef HAVE_TERMIOS_H
#include <termios.h>
#endif

#ifdef HAVE_STROPTS_H
#include <stropts.h>
#endif

#ifdef HAVE_SYS_TERMIO_H
#include <sys/termio.h>
#endif

#ifndef SIGCHLD
#define SIGCHLD	SIGCLD
#endif

static struct termios termorig;
static struct winsize winorig;

struct pst
{
  char *master;
  int mfd;
  char *slave;
  int sfd;
} pspair;

static char *progname;
char *shell;
char start_msg[BUFSIZ];
FILE *fscript;
FILE *fttime;
FILE *finput;

static void bye (int);
static void newwinsize (int);
static void prepchild (struct pst *);
static void rawmode (int);
static int findms (struct pst *);
void mysyslog (int, const char *, ...);

int
main (int argc, char *argv[], char *environ[])
{
  int n = 1;
  int valid = -1;
  unsigned char iobuf[BUFSIZ];
  char *p = NULL;
  char *user = getlogin ();
  char *sudo_user = NULL;
  char *vshell = NULL;
  char lscript[BUFSIZ];
  char ltime[BUFSIZ];
  char linput[BUFSIZ];
  time_t now;
  struct stat d, s;
  struct sigaction sa;
  struct timeval tv;
  double oldtime, newtime;
  pid_t pid = getpid ();

  progname = argv[0];

  if ((p = (char *) strrchr (progname, '/')) != NULL)
    progname = p + 1;


  if (argc >= 2)
    {
      if (!strcmp (argv[1], "-V") ||
	  !strcmp (argv[1], "-v") || !strcmp (argv[1], "--version"))
	{
	  fprintf (stdout, "%s version %s\n", progname, VERSION);
	  exit (0);
	}
      else if (!strcmp (argv[1], "-h") || !strcmp (argv[1], "--help"))
	{
	  fprintf (stdout,
		   "Usage: sudosh [REASON]\n"
		   "sudo shell that supports input and output logging to syslog\n"
		   "\n"
		   "-h, --help	display this help and exit\n"
		   "-v, --version	output version information and exit\n"
		   "\n" "Report bugs to <%s>\n", PACKAGE_BUGREPORT);
	  exit (0);
	}
      else
	{
	  fprintf (stderr, "sudosh: unrecognized option `%s'\n", *(argv + 1));
	  fprintf (stderr, "Try `sudosh --help' for more information.\n");
	  exit (1);
	}
    }

  if ((sudo_user = getenv ("SUDO_USER")) == NULL)
    {
      fprintf (stderr,
	       "sudosh must be called via sudo.  i.e. \"sudo sudosh\"\n");
      mysyslog (LOG_ERR,
		"sudosh must be called via sudo.  i.e. \"sudo sudosh\"");
      exit (255);
    }

  if (user == NULL)
    {
      fprintf (stderr,
	       "getlogin() returned NULL.  I do not know who you are.  Stopping.\n");
      mysyslog (LOG_ERR,
		"getlogin() returned NULL.  I do not know who you are.");
      exit (255);
    }

  if (stat ((const char *) LOGDIR, &d) == -1)
    {
      if (mkdir ((const char *) LOGDIR, S_IRWXU) == -1)
	{
	  char str[BUFSIZ];

	  snprintf (str, BUFSIZ - 1, "mkdir(%s)", LOGDIR);
	  perror (str);
	  fprintf (stderr,
		   "Directory %s needs to exist and be created by root with the permissions of 0700\n");
	  exit (errno);
	}
    }

  if ((shell = getenv ("SHELL")) == NULL)
    shell = getpwnam (user)->pw_shell;

  /* check against /etc/shells to make sure it's a real shell */
  setusershell ();
  while (vshell = (char *) getusershell ())
    {
      if (strcmp (shell, vshell) == 0)
	valid = 1;
    }
  endusershell ();

  if (valid != 1)
    {
      if (shell == NULL)
	{
	  fprintf (stderr, "Could not determine a valid shell.\n");
	  mysyslog (LOG_ERR, "Could not determine a valid shell");
	  exit (200);
	}
      else
	{
	  fprintf (stderr, "%s is not in /etc/shells\n", shell);
	  mysyslog (LOG_ERR, "%s,%s: %s is not in /etc/shells",
		    user, ttyname (0), shell);
	  exit (210);
	}
    }

  if (stat ((const char *) shell, &s) == -1)
    {
      fprintf (stderr, "Shell %s doesn't exist.\n", shell);
      mysyslog (LOG_ERR, "%s,%s: shell %s doesn't exist.",
		user, ttyname (0), shell);
      exit (255);
    }

  now = time ((time_t *) NULL);

  snprintf (lscript, (size_t) BUFSIZ - 1,
	    "%s/%s-script-%i", LOGDIR, user, now);
  snprintf (ltime, (size_t) BUFSIZ - 1, "%s/%s-time-%i", LOGDIR, user, now);
  snprintf (linput, (size_t) BUFSIZ - 1, "%s/%s-input-%i", LOGDIR, user, now);

  if ((fscript = fopen (lscript, "w")) == NULL)
    {
      fprintf (stderr, "fopen(): %s\n", lscript);
      perror ("fopen");
      exit (200);
    }
  if ((fttime = fopen (ltime, "w")) == NULL)
    {
      fprintf (stderr, "fopen(): %s\n", ltime);
      perror ("fopen");
      exit (200);
    }
  if ((finput = fopen (linput, "w")) == NULL)
    {
      fprintf (stderr, "fopen(): %s\n", linput);
      perror ("fopen");
      exit (200);
    }


  snprintf(start_msg, BUFSIZ-1, "starting session for %s,%s (%s) (%s-%i)",
	user, ttyname (0), shell, user, now);

  mysyslog (LOG_INFO, start_msg);
  mysyslog (LOG_INFO, "to view this session type: sudosh-replay %s-%i",
	    user, now);

  rawmode (0);

  if (findms (&pspair) < 0)
    {
      perror ("open pty failed");
      bye (1);
    }

  switch (fork ())
    {
    case 0:
      close (pspair.mfd);
      prepchild (&pspair);
    case -1:
      perror ("fork failed");
      bye (1);
    default:
      close (pspair.sfd);
    }

  setuid (getuid ());

  memset (&sa, 0, sizeof sa);
  sa.sa_handler = newwinsize;
  sa.sa_flags = SA_RESTART;
  sigaction (SIGWINCH, &sa, (struct sigaction *) 0);

  signal (SIGTERM, bye);
  signal (SIGCHLD, bye);

  oldtime = time (NULL);

  while (n > 0)
    {
      fd_set readfds;

      FD_ZERO (&readfds);
      FD_SET (pspair.mfd, &readfds);
      FD_SET (0, &readfds);

      gettimeofday ((struct timeval *) &tv, NULL);
      if (select (pspair.mfd + 1, &readfds, (fd_set *) 0,
		  (fd_set *) 0, (struct timeval *) 0) < 0)
	{

	  if (errno == EINTR)
	    continue;

	  perror ("select");
	  bye (1);
	}

      if (FD_ISSET (pspair.mfd, &readfds))
	{
	  if ((n = read (pspair.mfd, iobuf, sizeof (iobuf))) > 0)
	    {
	      if (write (1, iobuf, n) == -1)
		{
		  perror ("write stdout");
		  bye (1);
		}
	      if (fwrite (iobuf, 1, n, fscript) == -1)
		{
		  perror ("fwrite stdout");
		  bye (1);
		}
	    }
	  newtime = tv.tv_sec + (double) tv.tv_usec / 1000000;
	  fprintf (fttime, "%f %i\n", newtime - oldtime, n);
	  oldtime = newtime;
	}

      if (FD_ISSET (0, &readfds))
	{
	  if ((n = read (0, iobuf, BUFSIZ)) > 0)
	    {
	      if (write (pspair.mfd, iobuf, n) == -1)
		{
		  perror ("write master");
		  bye (1);
		}
	      switch (*iobuf)
		{
		case '\r':
		  fprintf (finput, "\n");
		  break;
		case 0x003:
		  fprintf (finput, "(CTRL-C)");
		  break;
		case 0x004:
		  fprintf (finput, "(CTRL-D)\n");
		  break;
		case 0x1a:
		  fprintf (finput, "(CTRL-Z)\n");
		  break;

		case 0x1b:
		  fprintf (finput, "(ESC)");
		  break;
		default:
		  (void) fwrite (iobuf, 1, n, finput);
		  break;
		}

	    }
	}
    }

  bye (0);
  return (0);
}

static int
findms (struct pst *p)
{
  char *sname;

  if ((p->mfd = open ("/dev/ptmx", O_RDWR)) == -1)
    {
      if ((p->mfd = open ("/dev/ptc", O_RDWR)) == -1)
	{
	  perror ("Cannot open cloning master pty");
	  return -1;
	}
    }

  (void) unlockpt (p->mfd);
  (void) grantpt (p->mfd);

  sname = (char *) ptsname (p->mfd);

  if ((p->sfd = open (sname, O_RDWR)) == -1)
    {
      perror ("open slave pty");
      close (p->mfd);
      return -1;
    }

  p->master = (char *) 0;

  p->slave = malloc (strlen (sname) + 1);
  strcpy (p->slave, sname);

#ifdef I_LIST
  if (ioctl (p->sfd, I_LIST, NULL) > 0)
    {
#ifdef I_FIND
      if (ioctl (p->sfd, I_FIND, "ldterm") != 1 &&
	  ioctl (p->sfd, I_FIND, "ldtty") != 1)
	{
#ifdef I_PUSH
	  (void) ioctl (p->sfd, I_PUSH, "ptem");
	  (void) ioctl (p->sfd, I_PUSH, "ldterm");
#endif
	}
#endif
    }
#endif

  return p->mfd;
}

static void
prepchild (struct pst *pst)
{
  int i;
  char *empty_env[] = { (char *) 0 };
  char *b = NULL;
  char *bshell = NULL;

  for (i = 0; i < 3; ++i)
    close (i);

  setsid ();

  if ((pst->sfd = open (pst->slave, O_RDWR)) < 0)
    exit (errno);

  dup (0);
  dup (0);

  for (i = 3; i < 100; ++i)
    close (i);

#ifdef TCSETS
  (void) ioctl (0, TCSETS, &termorig);
#endif
  (void) ioctl (0, TIOCSWINSZ, &winorig);

  setuid (getuid ());

  if ((b = (char *) strrchr (shell, '/')) != NULL)
    bshell = b + 1;

  fprintf(stderr, "%s\n", start_msg);
  execl(shell, bshell, (char *) 0);

  abort ();
}

static void
rawmode (int ttyfd)
{
  static struct termios termnew;

#ifdef TCGETS
  if (ioctl (ttyfd, TCGETS, &termorig) == -1)
    {
      perror ("ioctl TCGETS failed");
      exit (1);
    }
#endif

  if (ioctl (ttyfd, TIOCGWINSZ, &winorig) == -1)
    {
      perror ("ioctl TIOCGWINSZ failed");
      exit (1);
    }

  termnew.c_cc[VEOF] = 1;
  termnew.c_iflag = BRKINT | ISTRIP | IXON | IXANY;
  termnew.c_oflag = 0;
  termnew.c_cflag = termorig.c_cflag;
  termnew.c_lflag &= ~ECHO;

#ifdef TCSETS
  (void) ioctl (ttyfd, TCSETS, &termnew);
#endif
}

static void
bye (int signum)
{
#ifdef TCSETS
  (void) ioctl (0, TCSETS, &termorig);
#endif

  fclose (fttime);
  fclose (fscript);
  fclose (finput);

  exit (signum);
}

static void
newwinsize (int signum)
{
  int fd;

  if (ioctl (0, TIOCGWINSZ, &winorig) != -1)
    {
      if ((fd = open (pspair.slave, O_RDWR)) >= 0)
	{
	  (void) ioctl (fd, TIOCSWINSZ, &winorig);
	  close (fd);
	}
    }
}

void
mysyslog (int pri, const char *fmt, ...)
{
  va_list ap;
  char buf[BUFSIZ + 1];

  va_start (ap, fmt);
  vsnprintf (buf, sizeof (buf), fmt, ap);

  openlog (progname, 0, LOG_LOCAL2);
  syslog (pri, "%s", buf);
  closelog ();
}
